//
// Spikewave - this class implements a spike train generator
// 2001-12-27  Izumi Ohzawa, ohzawa@bpe.es.osaka-u.ac.jp
// 
// NOTE that the spike rate is not calibrated at all.  Displayed rate
// is NOT the actual mean discharge rate.  The numbers are strictly a
// ball-park ones.
// I merely use this to test the spike sorter and spike data acqusition system
// without any live subject.

// -------- Modified from the following ------------------
// a very simple Cocoa CoreAudio app
// by James McCartney  james@audiosynth.com  www.audiosynth.com
//
// Sinewave - this class implements a spike oscillator with dezippered control of frequency, pan and amplitude
//

#import "Spikewave.h"
#import <stdlib.h>

extern float expdev(long *idum);

// define a C struct from the Obj-C object so audio callback can access data
typedef struct {
    @defs(Spikewave);
} spikewavedef;

static long idum = -3;

static int wf[32] =
{ 0, 15, 50, 100, 180, 300, 490, 650, 800, 920, 980, 1000, 900, 630, 300, 0,
-300, -450, -530, -570, -520, -490, -390, -300, -200, -100, -20, 30, 80, 50, 20, 0
};

// this is the audio processing callback.
OSStatus appIOProc (AudioDeviceID  inDevice, const AudioTimeStamp*  inNow, const AudioBufferList*   inInputData, 
                             const AudioTimeStamp*  inInputTime, AudioBufferList*  outOutputData, const AudioTimeStamp* inOutputTime, 
                             void* defptr)
{    
    spikewavedef*	def = defptr; // get access to Spikewave's data
    int i;
    
    // load instance vars into registers
    double amp = def->amp;
    double pan = def->pan;
    double freq = def->freq;

    long cnt = def->cnt;
    long jitter=0, beg1=0, end1=0;
    float maxintvl;
	
    // numSamples comes out to 512 for 44.1kHz sampling rate ....
    int numSamples = def->deviceBufferSize / def->deviceFormat.mBytesPerFrame;
    
    // assume floats for now....
    float *out = outOutputData->mBuffers[0].mData;

    // Note that the spike rate is not calibrated at all.  Displayed rate
    // is NOT the actual mean discharge rate.
    maxintvl  = 1.0001/((freq+4)/44.0) ;

    if(cnt < 0) {
        jitter = (random() >> 10) & 511;		// 9 bit random number
        if(jitter > 480) jitter = 480;
        beg1 = jitter;
        end1 = beg1+32;
        cnt = expdev(&idum)*maxintvl;
    }
    
   
    for (i=0; i<numSamples; ++i) {
    
        float wave;
        
        if(cnt == 0 && i>= beg1 && i< end1)
            wave = amp * 0.001 * wf[i-beg1];
         else
            wave = 0;
 
        // write output
        *out++ = wave * (1.0-pan);		// left channel
        *out++ = wave * pan;			// right channel
        

    }
    
    cnt--;
   
    // save registers back to object
    def->freq = freq;
    def->amp = amp;
    def->pan = pan;
    def->cnt = cnt;
    return kAudioHardwareNoError;     
}


@implementation Spikewave

- (void)setup
{
    OSStatus				err = kAudioHardwareNoError;
    UInt32				count;    
    device = kAudioDeviceUnknown;
 
    initialized = NO;
   
    // get the default output device for the HAL
    count = sizeof(device);		// it is required to pass the size of the data to be returned
    err = AudioHardwareGetProperty(kAudioHardwarePropertyDefaultOutputDevice,  &count, (void *) &device);
    if (err != kAudioHardwareNoError) {
    	fprintf(stderr, "get kAudioHardwarePropertyDefaultOutputDevice error %ld\n", err);
        return;
    }
    
    // get the buffersize that the default device uses for IO
    count = sizeof(deviceBufferSize);	// it is required to pass the size of the data to be returned
    err = AudioDeviceGetProperty(device, 0, false, kAudioDevicePropertyBufferSize, &count, &deviceBufferSize);
    if (err != kAudioHardwareNoError) {
    	fprintf(stderr, "get kAudioDevicePropertyBufferSize error %ld\n", err);
        return;
    }
    fprintf(stderr, "deviceBufferSize = %ld\n", deviceBufferSize);
   
    // get a description of the data format used by the default device
    count = sizeof(deviceFormat);	// it is required to pass the size of the data to be returned
    err = AudioDeviceGetProperty(device, 0, false, kAudioDevicePropertyStreamFormat, &count, &deviceFormat);
    if (err != kAudioHardwareNoError) {
    	fprintf(stderr, "get kAudioDevicePropertyStreamFormat error %ld\n", err);
        return;
    }
    if (deviceFormat.mFormatID != kAudioFormatLinearPCM) {
    	fprintf(stderr, "mFormatID !=  kAudioFormatLinearPCM\n");
        return;
    }
    if (!(deviceFormat.mFormatFlags & kLinearPCMFormatFlagIsFloat)) {
    	fprintf(stderr, "Sorry, currently only works with float format....\n");
        return;
    }
    
    initialized = YES;

    fprintf(stderr, "--- Poisson Spike Gen audio API parameters ----\n");
    fprintf(stderr, "mSampleRate = %g\n", deviceFormat.mSampleRate);
    fprintf(stderr, "mFormatFlags = %08lX\n", deviceFormat.mFormatFlags);
    fprintf(stderr, "mBytesPerPacket = %ld\n", deviceFormat.mBytesPerPacket);
    fprintf(stderr, "mFramesPerPacket = %ld\n", deviceFormat.mFramesPerPacket);
    fprintf(stderr, "mChannelsPerFrame = %ld\n", deviceFormat.mChannelsPerFrame);
    fprintf(stderr, "mBytesPerFrame = %ld\n", deviceFormat.mBytesPerFrame);
    fprintf(stderr, "mNumSamples per call back = %ld\n",
                deviceBufferSize / deviceFormat.mBytesPerFrame);
    fprintf(stderr, "mBitsPerChannel = %ld\n", deviceFormat.mBitsPerChannel);
	
	amp=1.0;
	pan=0.5;
}


- (void)setAmpVal:(double)val
{
    amp = val;
}

- (void)setFreqVal:(double)val
{
    // freq = val * 2. * 3.14159265359 / deviceFormat.mSampleRate;
    freq = val;
    cnt = 0;
}

- (void)setPanVal:(double)val
{
    pan = val;
}


- (BOOL)start
{
    OSStatus		err = kAudioHardwareNoError;
   spikewavedef *def;

    if (!initialized) return false;
    if (soundPlaying) return false;
    
    // initialize phase and de-zipper filters.
    //phase = 0.0;
    cnt = 0;
//    rnd_init(12345);
    srandom(12345);
    expdev(&idum);

    def = (spikewavedef *)self;
    
    err = AudioDeviceAddIOProc(device, appIOProc, (void *) def);	// setup our device with an IO proc
    if (err != kAudioHardwareNoError) return false;
    
    err = AudioDeviceStart(device, appIOProc);				// start playing sound through the device
    if (err != kAudioHardwareNoError) return false;

    soundPlaying = true;						// set the playing status global to true
    return true;
}

- (BOOL)sstop
{
    OSStatus 	err = kAudioHardwareNoError;
    
    if (!initialized) return false;
    if (!soundPlaying) return false;
    
    err = AudioDeviceStop(device, appIOProc);				// stop playing sound through the device
    if (err != kAudioHardwareNoError) return false;

    err = AudioDeviceRemoveIOProc(device, appIOProc);			// remove the IO proc from the device
    if (err != kAudioHardwareNoError) return false;
    
    soundPlaying = false;						// set the playing status global to false
    return true;
}

@end
